#include "tripp.h"

#define FRAG_ERROR(x,...)print_error(0,"IP fragmenter: " ,x,## __VA_ARGS__)
#define FRAG_DEBUG(d,x,...)if(d)print_debug("IP fragmenter: " ,x,## __VA_ARGS__)

//from libnet1.0
int in_cksum(u_short *addr, int len){
   int sum;
   int nleft;
   u_short ans;
   u_short *w;
   
   sum = 0;
   ans = 0;
   nleft = len;
   w = addr;
   while (nleft > 1) {
   	sum += *w++;
	nleft -= 2;
	}
	if (nleft == 1) {
		*(u_char *)(&ans) = *(u_char *)w;
		sum += ans;
	}
	return (sum);
}


 

int compute_l4_cksum(struct ip_packet* pkt){

 #define CKSUM_CARRY(x) (x = (x >> 16) + (x & 0xffff), (~(x + (x >> 16)) & 0xffff))
 
  int cksum;
  int len;
  
  switch(pkt->ip->ip_p){
  
    case IPPROTO_TCP:{
      struct tcphdr* tcp=(struct tcphdr*) pkt->layer4;
      
        len=(LIBNET_TCP_H+pkt->layer4_options_size+pkt->payload_size);
  		tcp->th_sum=0;//it's probably already set to zero (by load_ip_packet)
  		cksum=in_cksum((u_short*) &(pkt->ip->ip_src),8);
  		cksum+=htons(IPPROTO_TCP+len);
  		cksum+=in_cksum((u_short*) tcp, len);
  		tcp->th_sum=CKSUM_CARRY(cksum);
  		
  		break;
  	}
  	
  	case IPPROTO_UDP:{
  	      struct udphdr* udp=(struct udphdr*) pkt->layer4;
      
        len=(LIBNET_UDP_H+pkt->payload_size);
  		udp->uh_sum=0;//it's probably already set to zero (by load_ip_packet)
  		cksum=in_cksum((u_short*) &(pkt->ip->ip_src),8);
  		cksum+=htons(IPPROTO_UDP+len);
  		cksum+=in_cksum((u_short*) udp, len);
  		udp->uh_sum=CKSUM_CARRY(cksum);
  		
  		break;
  	}
  	
  }
  
}




int fragment_packet(struct tqhead_p *pkl_h,struct ip_packet* pkt,u_int mtu,int linktype){

  struct ip_packet *frag;
  struct pktlist *fragcell;
  
  u_char tmpbuff[(MAXBUFFSIZE)*10];
  
  
  u_int	iphlen = LIBNET_IPV4_H + pkt->ip_options_size,

		l4hlen = pkt->headers_size - iphlen, 
  	
  		totsize = pkt->payload_size + l4hlen,
  
  		l2size = get_link_size(linktype),
  		
  		nfb = (mtu-iphlen)/8, off=0, fragcounter=1,
  		
  		frsize;
  
  
  if( (pkt->payload_size+pkt->headers_size) <= mtu){
  	fragcell=(struct pktlist *)malloc(sizeof(struct pktlist)); 
  	fragcell->p=pkt;	
  	TAILQ_INSERT_TAIL(pkl_h,fragcell,next);
  	return 1;
  }
  
  FRAG_DEBUG(DEBUG,"fragmenting %d bytes (payload+layer4 headers)",totsize);
  
  //non posso lasciare a libnet il compito di computare il cksum di layer4 perche 
  //questo sono pacchetti IP e libnet computerebbe il cksum per un singolo fragment
  //mentre a me serve il cksum di tutti i fragments riassemblati
  compute_l4_cksum(pkt);
  
  while(totsize>0){
 
		frag=(struct ip_packet *)malloc(sizeof(struct ip_packet));
		
		//last frag => append all remaining data
		if( totsize <= mtu)
			frsize = totsize;
		else
			frsize = (nfb*8);
		
		
		//first fragment => preserve layer4 headers
		if(off==0)
		{
			memcpy(tmpbuff, pkt->raw, iphlen + l2size + frsize);
			load_ip_packet(tmpbuff,frag,l2size+pkt->headers_size+frsize,linktype,1); 
			
		}
		else
		{
			//copy IP headers
			memcpy(tmpbuff, pkt->raw, iphlen + l2size);
			//skip layer4 headers
			memcpy(	tmpbuff + l2size + iphlen , pkt->layer4 + (off*8), frsize);
			//load_ip_packet(tmpbuff,frag,iphlen+frsize+l2size,linktype,	1);
			load_ip_packet(tmpbuff,frag,(pkt->headers_size-l4hlen)+frsize+l2size,linktype,	1);
		}			
		  
		  
		frag->payload_size=frsize;
		  
		//last frag => morefrag flag is set to zero
		if( totsize <= mtu)
		 frag->ip->ip_off =  htons(0 | (off) | frag->ip->ip_off );
		else 
		 frag->ip->ip_off =  htons(IP_MF | (off) | frag->ip->ip_off);
		   
		fragcell=(struct pktlist *)malloc(sizeof(struct pktlist)); 
		fragcell->p=frag;
		  
		  
		TAILQ_INSERT_TAIL(pkl_h,fragcell,next);  
		 
		off+=nfb; 
		totsize-=frsize;
		FRAG_DEBUG(DEBUG,"fragment %d size:%d",fragcounter,frag->payload_size);
		fragcounter++;
  }

  free(pkt);
  return 1;
}


