/* 	
	ipcsocket.c 	part of TRIPP Traffic Rewriter for IP (v4) Packets
	by poplix@papuasia.org 2005

*/


#include "tripp.h"
#include <sys/un.h>
#include <pwd.h>

#define IPC_ERROR(x,...)print_error(0,"IPC socket error: " ,x,## __VA_ARGS__)
#define IPC_DEBUG(d,x,...)if(d)print_debug("IPC: " ,x,## __VA_ARGS__)

#define SOCK_FILE ".tripp_socket"
    

  
struct ipcrule_ack{
	u_int32_t s_addr,
			  d_addr;
			  
	u_short sport,
			dport;
			
	u_int val;
	u_char opr;
};





int rmsocket(){
	char file[5000];
	struct passwd *pw;
		     	        
	pw = getpwuid(getuid());
	
	sprintf(file,"%s/%s",pw->pw_dir,SOCK_FILE);
	
	unlink(file);
	return 1;


}
  

    
int ipc_server(){
        int s, s2, t, len,done, n;
        struct sockaddr_un local, remote;
        char str[100];
		struct passwd *pw;
		     
		        
		pw = getpwuid(getuid());
		
        if ((s = socket(AF_UNIX, SOCK_STREAM, 0)) == -1) IPC_ERROR("socket error");
		
        
        local.sun_family = AF_UNIX;
        
        sprintf(local.sun_path,"%s/%s",pw->pw_dir,SOCK_FILE);
        //remove old socket if present
        unlink(local.sun_path);
        len = strlen(local.sun_path) + sizeof(local.sun_family) + 1;
        if (bind(s, (struct sockaddr *)&local, len) == -1) IPC_ERROR("bind error");
        
        if (listen(s, 5) == -1) IPC_ERROR("listen error");
    	
        //printf("Waiting for a connection...\n");
        t = sizeof(remote);
        if ((s2 = accept(s, (struct sockaddr *)&remote, &t)) == -1)IPC_ERROR("accept");
            

            //printf("Connected.\n");
		return s2;
            
 }
 
 
 int ipc_send(u_char* data,u_int size, int sok){
 
 	send(sok, data, size, 0);               

        return 1;
}


int ipc_recv(u_char *buff, u_int size,int sok){
	int n;
	fcntl(sok, F_SETFL, O_NONBLOCK);
	n = recv(sok, buff, size, 0);
    
    return n;
                
}


int ipc_sendrule_ack(u_int32_t s_addr,u_int32_t d_addr,u_short sport,u_short dport,u_char opr,u_int val,int sok){
	
	struct ipcrule_ack r;

	r.s_addr=s_addr;
	r.d_addr=d_addr;
	r.sport=sport;
	r.dport=dport;
	r.opr=opr;
	r.val=val;

	ipc_send((u_char*)&r,sizeof(struct ipcrule_ack),sok);

	return 1;
}



int ipc_getrule_ack(int sok,struct tqhead_f *tqp,struct tqhead_c *tqc){
	
	struct ipcrule_ack *r;
	
	u_char buff[1024];
	int n;
		//struct tqhead_f *tqp;
	struct filter *f;
	struct tcpchan *tcpc;
	

	memset(buff,0,sizeof(buff));
	
	n=ipc_recv(buff,sizeof(buff),sok);
	if(n<=0 && errno==EAGAIN)return 0;
	r=(struct ipcrule_ack*)buff;
	
	//printf("--->%s\n",int_to_ip(r->s_addr));
	
	tcpc=gettcpchan(r->s_addr,r->d_addr,ntohs(r->sport),ntohs(r->dport),tqc);
	
	if(tcpc==NULL){
		f=(struct filter *)malloc(sizeof(struct filter));
		if(f == NULL)exit(0);
					  
					  
		TAILQ_INIT((&(f->actionstq)));
		TAILQ_INIT((&(f->matchestq)));
		
		//r->dport=htons(r->dport);
		matchadd(TCP_DPORT,(u_char*)&(r->dport),2,OPERATOR_EXACT,0,(&(f->matchestq)));
		
		//r->sport=htons(r->sport);//sport first
		matchadd(TCP_SPORT,(u_char*)&(r->sport),2,OPERATOR_EXACT,0,(&(f->matchestq)));
		//r->s_addr=htonl(r->s_addr);
		matchadd(IP_SRC,(u_char*)(&(r->s_addr)),4,OPERATOR_EXACT,0,(&(f->matchestq)));
		//r->d_addr=htonl(r->d_addr);
		matchadd(IP_DST,(u_char*)(&(r->d_addr)),4,OPERATOR_EXACT,0,(&(f->matchestq)));

		
		//r->val=htonl(r->val);
		IPC_DEBUG(DEBUG,"setting value:%d\n",r->val);
		actionadd(TCP_ACK,(u_char*)(&(r->val)),4,NULL,0,0,r->opr,(&(f->actionstq)));//ack
		
		
		f->actcmd=ACT_SET;
		f->id=filter_nextindex(tqp);
		f->layer4=IPPROTO_TCP;
		TAILQ_INSERT_TAIL(tqp,f,next);
		
		tcpc=newtcpchan(r->s_addr,r->d_addr,ntohs(r->sport),ntohs(r->dport),f->id);
		TAILQ_INSERT_HEAD(tqc,tcpc,next);
		
	}else{

		struct filter* fn;
		struct actions* modact;

		fn=getrule(tcpc->ruleid,tqp);
		if(fn==NULL){printf("eror\n");exit(9);}
		
		//ruleack ha 4 match e un solo set...
		modact=TAILQ_FIRST(&(fn->actionstq));
		
		IPC_DEBUG(DEBUG,"modify existing rule  val:%d  opr:%d\n",r->val,r->opr);
		memcpy(modact->rplc,(&(r->val)),4);
		modact->opr=r->opr;
		
		
	}
	return n;	
	
}
    
    
    
int ipc_connect(){
        int s, t, len;
        struct sockaddr_un remote;
        char str[100];
		struct passwd *pw;
		pw = getpwuid(getuid());
        
        if ((s = socket(AF_UNIX, SOCK_STREAM, 0)) == -1) IPC_ERROR("socket");
        

        remote.sun_family = AF_UNIX;
        sprintf(remote.sun_path,"%s/%s",pw->pw_dir,SOCK_FILE);
        
        len = strlen(remote.sun_path) + sizeof(remote.sun_family)+1;
        
        if (connect(s, (struct sockaddr *)&remote, len) == -1) IPC_ERROR("connect");


        return s;
}



