/* 	
	replacers.c 	part of TRIPP Traffic Rewriter for IP (v4) Packets
	by poplix@papuasia.org 2005

*/

#include "tripp.h"


#define FILTER_ERROR(x,...)print_error(0,"filter error: " ,x,## __VA_ARGS__)
#define FILTER_DEBUG(d,x,...)if(d)print_debug("filter:" ,x,## __VA_ARGS__)


int replace_header(struct ip_packet *pkt, struct tqhead_a *actshead,struct tqhead_f *);


char *meminmem(char *str,u_int str_size,char *buff,u_int buff_size);

u_char filter_getl4(u_char ft){
	if(0<ft<6)return IPPROTO_IP;
	if(5<ft<13)return IPPROTO_TCP;
	if(12<ft<15)return IPPROTO_UDP;
}

int adjustsynack(int delta,u_int32_t saddr,u_int32_t daddr,u_short sport,u_short dport,int ipc_sok,struct tqhead_f *tqp,struct tqhead_c *tqc);

//confronta un ip_packet con una lista di condizioni
//nota che tutti i confronti e le assegnazioni vengono fatte a livello di memoria.
//per ogni field da confrontare/settare e' memorizza la dimensione in bytes,dunque qualunque valore (int,short,char)
//viene considerato come stream di bytes
int match_packet(struct ip_packet *pkt,struct tqhead_m *matchh){
	//nota che non posso passare memcmp direttamente a LXOR senno verrebbe richiamata due volte
	#define LXOR(x,y) ( (x || y) && (!(x && y)) )
	
	struct matches *match;

	u_char l4t;
	int cmp;
	//scorre la lista delle condizioni
	
	for (match = TAILQ_FIRST(matchh); match != NULL; match = TAILQ_NEXT(match, next)){
		l4t=filter_getl4(match->type);
		//se match ip processa altrimenti continue se il paccketto non ha lo stesso l4 della condizione di match
		if(match->type!=PAYLOAD && l4t!=IPPROTO_IP && pkt->ip->ip_p!=l4t)FILTER_ERROR("error1");
		
		
		//in base al tipo di field applica un memcpy tra il val del field e il val della condizione
		switch(match->type){
			case IPTOS:
				cmp=(!memcmp((u_char*)(&(pkt->ip->ip_tos)),match->value,match->value_size));
				if(!LXOR(match->opr_not,cmp))return 0;
					
				FILTER_DEBUG(DEBUG,"pkt_match TOS");
				
				break;
		
			case IP_ID:
				cmp=(!memcmp((u_char*)(&(pkt->ip->ip_id)),match->value,match->value_size));
				if(!LXOR(match->opr_not,cmp))return 0;
				
				FILTER_DEBUG(DEBUG,"pkt_match ID");
			
				break;
			case IPTTL:
				cmp=(!memcmp((u_char*)(&(pkt->ip->ip_ttl)),match->value,match->value_size));
				if(!LXOR(match->opr_not,cmp))return 0;
				FILTER_DEBUG(DEBUG,"pkt_match TTL");
				
				break;
			case IP_SRC:
				cmp=(!memcmp((u_char*)(&(pkt->ip->ip_src.s_addr)),match->value,match->value_size));
				if(!LXOR(match->opr_not,cmp))return 0;
					
				FILTER_DEBUG(DEBUG,"pkt_match IP SRC");
				
				break;
			case IP_DST:
				cmp=(!memcmp((u_char*)(&(pkt->ip->ip_dst.s_addr)),match->value,match->value_size));
				if(!LXOR(match->opr_not,cmp))return 0;
				FILTER_DEBUG(DEBUG,"pkt_match IP DST");
				break;
				
				
			case IP_FLAGS_MF:
				cmp=( ( pkt->ip->ip_off & IP_MF ));
				if(!LXOR(match->opr_not,cmp))return 0;
				FILTER_DEBUG(DEBUG,"pkt_match IP  FLAGS MF");
				break;	
			case IP_FLAGS_DF:
				cmp=( ( pkt->ip->ip_off & IP_DF ));
				if(!LXOR(match->opr_not,cmp))return 0;
				FILTER_DEBUG(DEBUG,"pkt_match IP FLAGS DF");
				break;
			case IP_FLAGS_RF:
				cmp=( ( pkt->ip->ip_off & IP_RF ));
				if(!LXOR(match->opr_not,cmp))return 0;
				FILTER_DEBUG(DEBUG,"pkt_match IP FLAGS RF");
				break;
				
				
				
			case TCP_SPORT:
				cmp=(!memcmp((u_char*)(&((struct tcphdr*)pkt->layer4)->th_sport),match->value,match->value_size));
				if(!LXOR(match->opr_not,cmp))return 0;
					
				FILTER_DEBUG(DEBUG,"pkt_match TCP SPORT");
				
				break;
				
			case TCP_DPORT:
				cmp=(!memcmp((u_char*)(&((struct tcphdr*)pkt->layer4)->th_dport),match->value,match->value_size));
				if(!LXOR(match->opr_not,cmp))return 0;
					
				FILTER_DEBUG(DEBUG,"pkt_match TCP DPORT");
				
				break;
			case TCP_SEQ:
				cmp=(!memcmp((u_char*)(&((struct tcphdr*)pkt->layer4)->th_seq),match->value,match->value_size));
				if(!LXOR(match->opr_not,cmp))return 0;

				FILTER_DEBUG(DEBUG,"pkt_match SEQ");
				
				break;
			case TCP_ACK:
				cmp=(!memcmp((u_char*)(&((struct tcphdr*)pkt->layer4)->th_ack),match->value,match->value_size));
				if(!LXOR(match->opr_not,cmp))return 0;

				FILTER_DEBUG(DEBUG,"pkt_match ACK");
				
				break;
			case TCP_FLAGS:
				if(match->opr==OPERATOR_EXACT){
				  cmp=(!memcmp((u_char*)(&((struct tcphdr*)pkt->layer4)->th_flags),match->value,match->value_size));
				  if(!LXOR(match->opr_not,cmp))return 0;
					
			      FILTER_DEBUG(DEBUG,"pkt_match flags EXACT %d\n",(u_int)(match->opr_not) );
				 } 
				
				if(match->opr==OPERATOR_CONTAINS){
				//l' ==*(match->value) serve perche senno se chiedo syn+ack lui mi match tutti i pk che hanoo syn
				// e tutti quelli che hanno ack,, invece io voglio che mi match solo quelli che hanno sia syn che ack
				  cmp=( ( (((struct tcphdr*)pkt->layer4)->th_flags) & *(match->value) )==*(match->value));
				  if(!LXOR(match->opr_not,cmp))return 0;
					
					FILTER_DEBUG(DEBUG,"pkt_match flags CONTAINS");
				  
				}
				break;
			case TCP_WIN:
				cmp=(!memcmp((u_char*)(&((struct tcphdr*)pkt->layer4)->th_win),match->value,match->value_size));
				if(!LXOR(match->opr_not,cmp))return 0;
					
				FILTER_DEBUG(DEBUG,"pkt_match WIN");
				
				break;
			case TCP_URP:
				cmp=(!memcmp((u_char*)(&((struct tcphdr*)pkt->layer4)->th_urp),match->value,match->value_size));
				if(!LXOR(match->opr_not,cmp))return 0;
					
				FILTER_DEBUG(DEBUG,"pkt_match URP");
				
				break;
				
				
			case TCP_OPTIONS:
				cmp=(!memcmp(pkt->layer4_options,match->value,match->value_size));
				if(!LXOR(match->opr_not,cmp))return 0;
					
				FILTER_DEBUG(DEBUG,"pkt_match TCP OPTIONS");
				
				break;
				
				
			case UDP_SPORT:
				cmp=(!memcmp((u_char*)(&((struct udphdr*)pkt->layer4)->uh_sport),match->value,match->value_size));
				if(!LXOR(match->opr_not,cmp))return 0;

				FILTER_DEBUG(DEBUG,"pkt_match UDP SPORT");
				
				break;
			case UDP_DPORT:
				cmp=(!memcmp((u_char*)(&((struct udphdr*)pkt->layer4)->uh_dport),match->value,match->value_size));
				if(!LXOR(match->opr_not,cmp))return 0;
					
				FILTER_DEBUG(DEBUG,"pkt_match UDP DPORT");
				
				break;
			case PAYLOAD:
		  		if(pkt->payload_size==0){
		  			FILTER_DEBUG(DEBUG,"trying to match a packet with empty payload..skipping..");
		  			break;
		  		}
				if(match->opr==OPERATOR_EXACT){
				  //cmp=(!memcmp(pkt->payload,match->value,match->value_size));
				  cmp=(!memcmp(pkt->payload,match->value,pkt->payload_size));
				  if(!LXOR(match->opr_not,cmp))return 0;
					
					
					FILTER_DEBUG(DEBUG,"pkt_match PAYLOAD EXACT %d",(u_int)(match->opr_not) );
				  
				}
				if(match->opr==OPERATOR_CONTAINS){
				  cmp=( (meminmem(match->value,match->value_size,pkt->payload,pkt->payload_size)) != NULL );
				  if(!LXOR(match->opr_not,cmp))return 0;
					
					FILTER_DEBUG(DEBUG,"pkt_match PAYLOAD CONTAINS");
				  
				}
				break;

		  }
		//if(matched==0)return 0;
	}
		
	return 1;
}












struct tcpchan *newtcpchan(u_int32_t saddr,u_int32_t daddr,u_short sport,u_short dport,u_int ruleid){
	struct tcpchan *tc;
	
	tc=(struct tcpchan*)malloc(sizeof(struct tcpchan));
	if(tc==NULL)exit(0);
	
	tc->saddr=saddr;
	tc->daddr=daddr;
	tc->sport=sport;
	tc->dport=dport;
	tc->ruleid=ruleid;
	
	return tc;
	
}


struct tcpchan *gettcpchan(u_int32_t saddr,u_int32_t daddr,u_short sport,u_short dport,struct tqhead_c *tqc){
	struct tcpchan *fp;
	
		for (fp = TAILQ_FIRST(tqc); fp != NULL; fp = TAILQ_NEXT(fp, next))
			if(	fp->saddr==saddr	&&
				fp->daddr==daddr	&&
				fp->sport==sport	&&
				fp->dport==dport	)
							return fp;
							
		return NULL;
}



struct filter *getrule(u_int index,struct tqhead_f* tqp){
	struct filter *fp;
	//struct tqhead_f *tqp;

	//tqp = &filtertq;	
	
	for (fp = TAILQ_FIRST(tqp); fp != NULL; fp = TAILQ_NEXT(fp, next))
		if(fp->id==index)return fp;
		
	return NULL;
	
}






//scorre la lista delle regole e le applica all' ip_packet
int process_packet(struct ip_packet *pkt,int ipc_sok,struct tqhead_f* tqp,struct tqhead_c* tqc){
	struct filter *fp;
	int psdelta=0;
	for (fp = TAILQ_FIRST(tqp); fp != NULL; fp = TAILQ_NEXT(fp, next)){
	  if(fp->layer4==0 || pkt->ip->ip_p==fp->layer4)
		switch(fp->actcmd){
		  case ACT_PREWRITE:
		  case ACT_SET:
			if(match_packet(pkt, (&(fp->matchestq))) )psdelta+=replace_header(pkt, (&(fp->actionstq)),tqp);
			break;
		  case ACT_CLONE:
		  	if( match_packet(pkt, (&(fp->matchestq))) )pkt->counter=fp->intval;
		  case ACT_FRAGMENT:
		  	if( match_packet(pkt, (&(fp->matchestq))) )pkt->fragsize=fp->intval;
		  	break;
		 case ACT_DROP:
		  	if( match_packet(pkt, (&(fp->matchestq))) )pkt->counter=0;
		  	break;
		
		}
	}

	if(pkt->ip->ip_p == IPPROTO_TCP && psdelta!=0){
		//mantieni la coerenza tra syn e ack una volta modificata la dimensione del payload
		//questa operaz usa un processo parallelo per modificare gli ack e ingannare lo stack locale
		adjustsynack(psdelta,
							pkt->ip->ip_src.s_addr,pkt->ip->ip_dst.s_addr,
							((struct tcphdr*)pkt->layer4)->th_sport,
							((struct tcphdr*)pkt->layer4)->th_dport,
							ipc_sok,tqp,tqc);
	}
}


int modifyhfield(u_char* fp,u_char* rplc, u_int size,u_char opr){
	
	u_long lval;
	u_short sval;
	u_char cval;
	
	//u_char values
	if(size==1){
		if(opr==OPERATOR_INCREASE){
		  cval=( (*fp) +  (*rplc) );
		  memcpy((u_char*)fp,(u_char*)&cval,size);
		}
		if(opr==OPERATOR_DECREASE){
		  cval=( (*fp) - (*rplc) );
		  memcpy((u_char*)fp,(u_char*)&cval,size);
		}
		if(opr==OPERATOR_EXACT)
			memcpy((u_char*)fp,(u_char*)rplc,size);
	}
	
	//u_short values
	if(size==2){
		if(opr==OPERATOR_INCREASE){
		  sval=( ntohs( *(u_short*)fp ) + ntohs( *(u_short*)rplc ));
		  sval=htons(sval);
		  memcpy((u_char*)fp,(u_char*)&sval,size);
		}
		if(opr==OPERATOR_DECREASE){
		  sval=( ntohs( *(u_short*)fp ) - ntohs( *(u_short*)rplc ));
		  sval=htons(sval);
		  memcpy((u_char*)fp,(u_char*)&sval,size);
		}
		if(opr==OPERATOR_EXACT)
			memcpy((u_char*)fp,(u_char*)rplc,size);
	}
	
	//u_long values
	if(size==4){
		if(opr==OPERATOR_INCREASE){
		  lval=( ntohl( *(u_int*)fp ) + ntohl( *(u_int*)rplc ));
		  lval=htonl(lval);
		  memcpy((u_char*)fp,(u_char*)&lval,size);
		}
		if(opr==OPERATOR_DECREASE){
		  lval=( ntohl( *(u_int*)fp ) - ntohl( *(u_int*)rplc ));
		  lval=htonl(lval);
		  memcpy((u_char*)fp,(u_char*)&lval,size);
		}
		if(opr==OPERATOR_EXACT)
			memcpy((u_char*)fp,(u_char*)rplc,size);
	}	

return 1;			
}




//simile a match_pachet.. socrre la lista delle azioni ed applica le sostituzioni su ip_packet
//restituisce la differenza di payload_size
int replace_header(struct ip_packet *pkt, struct tqhead_a *actshead,struct tqhead_f *tqp){
	struct actions *act;
	u_char l4t;
	int payload_replacements=0;
	int payload_size_delta=0;
	
	
	//struct tqhead_f *tqp;
	struct filter *f,*fn;
	
	struct tcpchan *tcpc;
	
	u_char opr;
	
	u_short sval;
	u_int 	lval;

	struct actions *modact;
	u_char newopr;
	int newval; //puo essere negativo...
	
	int psdelta=0;
	
	int a;//tmp
	
	for (act = TAILQ_FIRST(actshead); act != NULL; act = TAILQ_NEXT(act, next)){
		l4t=filter_getl4(act->rtype);
		//se match ip processa altrimenti continue se il paccketto non ha lo stesso l4 della condizione di match
		if(l4t!=0 && pkt->ip->ip_p!=l4t)FILTER_ERROR("error2");
	
		switch(act->rtype){
		  case IPTOS:
			modifyhfield( (u_char*)(&(pkt->ip->ip_tos)), (u_char*)act->rplc,act->rplc_size,act->opr);
			FILTER_DEBUG(DEBUG,"pkt_set TOS");
			break;
		  case IP_ID:
		    modifyhfield( (u_char*)(&(pkt->ip->ip_id)), (u_char*)act->rplc,act->rplc_size,act->opr);
		    FILTER_DEBUG(DEBUG,"pkt_set ID");
			break;
		  case IPTTL:
		    modifyhfield( (u_char*)(&(pkt->ip->ip_ttl)), (u_char*)act->rplc,act->rplc_size,act->opr);
		    FILTER_DEBUG(DEBUG,"pkt_set TTL");
			break;
		  case IP_SRC:
		    modifyhfield( (u_char*)(&(pkt->ip->ip_src.s_addr)), (u_char*)act->rplc,act->rplc_size,act->opr);
		    FILTER_DEBUG(DEBUG,"pkt_set IP SRC");
			break;
		  case IP_DST:
		    modifyhfield( (u_char*)(&(pkt->ip->ip_dst.s_addr)), (u_char*)act->rplc,act->rplc_size,act->opr);			
		    FILTER_DEBUG(DEBUG,"pkt_set IP DST");
		    break;
		  case IP_FLAGS_MF:
		    if(*act->rplc == 1)pkt->ip->ip_off=htons(IP_MF | pkt->ip->ip_off);
		    if(*act->rplc == 0)pkt->ip->ip_off=\
		    	htons( pkt->ip->ip_off & (pkt->ip->ip_off ^ IP_MF));	
		    FILTER_DEBUG(DEBUG,"pkt_set IP_MF");
			break;
		  case IP_FLAGS_DF:
		    if(*act->rplc == 1)pkt->ip->ip_off=htons(IP_DF | pkt->ip->ip_off);
		    if(*act->rplc == 0)pkt->ip->ip_off=\
		    	htons( pkt->ip->ip_off & (pkt->ip->ip_off ^ IP_DF));
		    FILTER_DEBUG(DEBUG,"pkt_set IP_DF");
			break;
		  case IP_FLAGS_RF:
		    if(*act->rplc == 1)pkt->ip->ip_off=htons(IP_RF | pkt->ip->ip_off);
		    if(*act->rplc == 0)pkt->ip->ip_off=\
		    	htons( pkt->ip->ip_off & (pkt->ip->ip_off ^ IP_RF));
		    FILTER_DEBUG(DEBUG,"pkt_set IP_RF");
			break;
		  case TCP_SPORT:
			modifyhfield( (u_char*)(&(((struct tcphdr*)pkt->layer4)->th_sport)), (u_char*)act->rplc,act->rplc_size,act->opr);
			FILTER_DEBUG(DEBUG,"pkt_set TCP SPORT");
			break;
		  case TCP_DPORT:
		  	modifyhfield( (u_char*)(&(((struct tcphdr*)pkt->layer4)->th_dport)), (u_char*)act->rplc,act->rplc_size,act->opr);
		  	FILTER_DEBUG(DEBUG,"pkt_set TCP DPORT");
			break;
			
		  case TCP_SEQ:
			modifyhfield( (u_char*)(&(((struct tcphdr*)pkt->layer4)->th_seq)), (u_char*)act->rplc,act->rplc_size,act->opr);
			FILTER_DEBUG(DEBUG,"pkt_set SEQ");
			break;
			
		  case TCP_ACK:
		    modifyhfield( (u_char*)(&(((struct tcphdr*)pkt->layer4)->th_ack)), (u_char*)act->rplc,act->rplc_size,act->opr);
		    FILTER_DEBUG(DEBUG,"pkt_set ACK");
			break;
			
			
		  case TCP_FLAGS:
		  	if(act->opr==OPERATOR_EXACT)
		  		memcpy((u_char*)(&((struct tcphdr*)pkt->layer4)->th_flags),act->rplc,act->rplc_size);
			if(act->opr==OPERATOR_INCREASE)
				((struct tcphdr*)pkt->layer4)->th_flags |= (*((u_char*)act->rplc));
			if(act->opr==OPERATOR_DECREASE)
				((struct tcphdr*)pkt->layer4)->th_flags &= \
					( (*((u_char*)act->rplc)) ^ ((struct tcphdr*)pkt->layer4)->th_flags );
			
			FILTER_DEBUG(DEBUG,"pkt_set TCP FLAGS");
			break;
			
		  case TCP_WIN:
		    modifyhfield((u_char*)(&((struct tcphdr*)pkt->layer4)->th_win), (u_char*)act->rplc,act->rplc_size,act->opr);
		    FILTER_DEBUG(DEBUG,"pkt_set WIN");
			break;
		  case TCP_URP:
		    modifyhfield((u_char*)(&((struct tcphdr*)pkt->layer4)->th_urp), (u_char*)act->rplc,act->rplc_size,act->opr);
		    FILTER_DEBUG(DEBUG,"pkt_set URP");
			break;
			
		  case UDP_SPORT:
		    modifyhfield((u_char*)(&((struct udphdr*)pkt->layer4)->uh_sport), (u_char*)act->rplc,act->rplc_size,act->opr);
		    FILTER_DEBUG(DEBUG,"pkt_set UDP SPORT");
			break;
		  case UDP_DPORT:
  		    modifyhfield((u_char*)(&((struct udphdr*)pkt->layer4)->uh_dport), (u_char*)act->rplc,act->rplc_size,act->opr);
  		    FILTER_DEBUG(DEBUG,"pkt_set UDP DPORT");
			break;
			
		  case PAYLOAD:
	 		if(pkt->payload_size==0){
		  		FILTER_DEBUG(DEBUG,"try to set payload to a packet with no payload ..skipping..");
		  		break;
		  	}
		  	payload_size_delta=pkt->payload_size;
		  	memcpy(pkt->payload,act->rplc,act->rplc_size);
		  	pkt->payload_size=act->rplc_size;
			payload_size_delta-=pkt->payload_size;
			
			FILTER_DEBUG(DEBUG,"pkt_set PAYLOAD");
			break;
			
		  case PAYLOAD_REWRITE:
		  	if(pkt->payload_size==0){
		  		FILTER_DEBUG(DEBUG,"try to rewrite payload to a packet with no payload ..skipping..");
		  		//break;
		  		return 0; //psdelta=0;()
		  	}
		  	payload_size_delta=pkt->payload_size;
		  	payload_replacements = memreplace(act->matchpayload,act->matchpayload_size,act->rplc,act->rplc_size,pkt->payload,pkt->payload_size,act->replacetimes);
			pkt->payload_size += ( payload_replacements * (act->rplc_size - act->matchpayload_size) );
			payload_size_delta-=pkt->payload_size;

			FILTER_DEBUG(DEBUG,"pkt_REWRITE PAYLOAD");
			break;
			
		  case TCP_OPTIONS:
		  	memcpy(pkt->layer4_options,act->rplc,act->rplc_size);
		  	pkt->layer4_options_size=act->rplc_size;
		  break;
			
		}
		psdelta+=payload_size_delta;
	}
	
	return psdelta;
}



char *meminmem(char *str,u_int str_size,char *buff,u_int buff_size){
    char *off=NULL;    
    
    int a;
    
    if(buff_size == 0) return NULL;
    
    //find offset per la prima occorrenza di str
   	for(a=0;a<buff_size;a++)
		if(buff[a]==str[0])
			if(!memcmp(str,(buff+a),str_size)){
				off = ( buff + a);
				break;
			}
	
    return off;
 }



//sostituisce una str con un'altra un un buffer
int memreplace(char *str, u_int str_size, char* rplc, u_int rplc_size, char *buff, u_int buff_size,int limit){
    char *off=NULL;

    char *end= (buff+buff_size);
    int len,cnt=0;
    
    if(buff_size == 0) return 0;
    
    //find offset per la prima occorrenza di str

	off=meminmem(str,str_size,buff,buff_size);
    if(off == NULL)return 0;
    
    //fai spazio nel buffer per il replacement (che puo essere piu grande o piu piccolo di str)
    len = end - off;  
    memmove( off + rplc_size, off + str_size, len); 
    memcpy(off,rplc, rplc_size);
	//cnt++;
	//se il buffer non e' finito e il limite de replc imposto non e' raggiunto applica la ricorsione
	limit--;
    if(len+(rplc_size-str_size) >= str_size && limit!=0 )
		//cnt += memreplace(str, str_size, rplc, rplc_size, off + rplc_size, len,limit);		
		cnt += memreplace(str, str_size, rplc, rplc_size, off + rplc_size,len+((rplc_size-str_size)) ,limit);		
	
	//ritorna il contatore tel tot dei rplc
	return ++cnt;
 }
 

 



int adjustsynack(int delta,u_int32_t saddr,u_int32_t daddr,u_short sport,u_short dport,int ipc_sok,struct tqhead_f *tqp,struct tqhead_c *tqc){
	//puntatore al nuovo filtro (o da modificare)
	struct filter *f;
	//tcp channel
	struct tcpchan *tcpc;
	
	//short and long values (probably unusefoul)
	u_short sval;
	u_int 	lval;
	//azione a cui modificare il valore (che e' il delta)
	struct actions *modact;
	
	//new opr,val per modificare l'azione (modact)
	u_char newopr;
	int newval; //puo essere negativo...
	//segni dei valori assoluti calcolati in base all'operatore (INC/DEC)
	int s1,s2;
	
	//absolute value of delta and his operator
	u_int payload_size_delta;
	u_char opr;
	
	//compute abs value
	if(delta < 0){payload_size_delta=(delta * (-1)); opr=OPERATOR_INCREASE;}
	else  {payload_size_delta=delta; opr=OPERATOR_DECREASE;}
	
	FILTER_DEBUG(DEBUG,"delta:%d pd:%d opr:%d",delta,payload_size_delta,opr);
	
	tcpc=gettcpchan(saddr,daddr,ntohs(sport),ntohs(dport),tqc);
	//
	//se e' la prima volta che si modifica il payload a questo canale
	//aggiungi la regola e registra il nuovo canale
	if(tcpc==NULL){
		f=(struct filter *)malloc(sizeof(struct filter));
		if(f == NULL){printf("malloc error\n");exit(0);}

		 //mancano htonl
	 
		 TAILQ_INIT((&(f->actionstq)));
		 TAILQ_INIT((&(f->matchestq)));




		sval=sport;
		matchadd(TCP_SPORT,(u_char*)&sval,2,OPERATOR_EXACT,0,(&(f->matchestq)));
		
		sval=dport;
		matchadd(TCP_DPORT,(u_char*)&sval,2,OPERATOR_EXACT,0,(&(f->matchestq)));		
		
		//printf("svo1\n");
		lval=daddr;
		matchadd(IP_DST,(u_char*)&lval,4,OPERATOR_EXACT,0,(&(f->matchestq)));
		//printf("svo2 %s\n",int_to_ip(saddr));
		
		lval=saddr;//unusefoul (mind)
		matchadd(IP_SRC,(u_char*)&lval,4,OPERATOR_EXACT,0,(&(f->matchestq)));
		
		//printf("svo3\n");

		
		FILTER_DEBUG(DEBUG,"setting value:%d\n",payload_size_delta);
		payload_size_delta=htonl(payload_size_delta);
		actionadd(TCP_SEQ,(u_char*)&payload_size_delta,4,NULL,0,0,opr,(&(f->actionstq)));//seq
		
		FILTER_DEBUG(DEBUG,"mask rule added\n");
		f->actcmd=ACT_SET;
		f->id=filter_nextindex(tqp);
		f->layer4=IPPROTO_TCP;
		//insert head perche senno se inserisco in coda la regola viene applicata
		//in questo setesso ciclo, indi modifica seq del pkt con payload replaced
		TAILQ_INSERT_HEAD(tqp,f,next);
		
		//register new channel
		tcpc=newtcpchan(saddr,daddr,ntohs(sport),ntohs(dport),f->id);
		TAILQ_INSERT_HEAD(tqc,tcpc,next);	

	}
	//altrimenti modifica la regola gia esistente con il nuovo delta
	else {
		
		//puntatore alla regola da modificare (quella con l'id corrispondente al channel)
		f=getrule(tcpc->ruleid,tqp);
		
		if(f==NULL){printf("eror\n");exit(9);}
		//ruleack ha 4 match e un solo set...
		modact=TAILQ_FIRST(&(f->actionstq));
		newopr=modact->opr;
		newval=ntohl(*(u_int*)modact->rplc);
		
		if(opr==OPERATOR_INCREASE)s1=-1; else s1=1;
		if(newopr==OPERATOR_INCREASE)s2=-1; else s2=1;
		//compute new value
		newval=( (s1*payload_size_delta) + (s2*newval));
		//printf("-->%d\n",newval);
		
		//make it absolute 
		if(newval < 0){newval*=(-1); opr=OPERATOR_INCREASE;}
		else  opr=OPERATOR_DECREASE;
		
		//prepare delta for IPC sending BEFORE htonl
		//ipc expects addresses in NET ORDER and parameters in HOST ORDER
		
		
		//values in rules table are in NETWORK ORDER
		newval=htonl(newval);
		
		payload_size_delta=newval;
		memcpy(modact->rplc,&newval,4);
		modact->opr=opr;
		


			
	}
		
		//invia il delta e il canale al processo parallelo
		if(opr==OPERATOR_INCREASE)opr=OPERATOR_DECREASE;
		else opr=OPERATOR_INCREASE;
		
		ipc_sendrule_ack(daddr,saddr,dport,sport,opr,payload_size_delta,ipc_sok);
							
				
}


