/* 	
	tripp.c 	part of TRIPP Traffic Rewriter for IP (v4) Packets
	by poplix@papuasia.org 2005

     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
     
*/

#include "tripp.h"

#define INIT_ERROR(x,...)print_error(0,"init error: " ,x,## __VA_ARGS__)
#define RT_ERROR(x,...)print_error(0,"runtime error: " ,x,## __VA_ARGS__)
#define TRIPP_DEBUG(d,x,...)if(d)print_debug("" ,x,## __VA_ARGS__)


#ifdef BSD
#define DEFAULT_INIF "lo0"
#endif

#ifdef __linux__
#define DEFAULT_INIF "lo"
#endif


struct thparam{
	int singproc;
	struct tunnel *tun;
	char 	frules[1024];
	u_int32_t gw;
	
};



pid_t pid1;
pid_t pid2=0;
	
void* tripp_process(void *par);
	
	
int get_link_size(int dl){
	   switch (dl) {    
        case 	DLT_EN10MB:	return sizeof(struct ether_header);
        case  	DLT_LOOP:	return 4;
        default:return -1;
        }
 }
 
 

 
 
 int init_intfs(char *inif_name,char* outif_name,struct tunnel* tun){
	//defaults
	tun->inif.oldmtu=0;
	tun->outif.oldmtu=0;
	
	//getinfo
	if(!getifinfo(outif_name,&(tun->outif)) ||
       !getifinfo(inif_name,&(tun->inif)) ){
  			INIT_ERROR("error fetching interface infos\n");
  			
  	}

	if(tun->outif.ipaddr==0x00)
		INIT_ERROR("no ip for writeif\n");
		
	
	//adjust mtus
	if(tun->inif.mtu > tun->outif.mtu){
		setmtu(&(tun->inif),tun->outif.mtu);
		tun->mtu=tun->outif.mtu;
	}

	if(tun->inif.mtu < tun->outif.mtu){
		setmtu(&(tun->outif),tun->inif.mtu);
		tun->mtu=tun->inif.mtu;
	}
	
	if(tun->inif.mtu == tun->outif.mtu)
		tun->mtu=tun->inif.mtu;
	
	
	return 1;
}
 
 /*
int get_layer4_hdrsize(u_char proto){
	switch (proto) {
		case IPPROTO_TCP:  return LIBNET_TCP_H;
 		case IPPROTO_UDP:  return LIBNET_UDP_H;
 			
 		default: return -1;
 	}
}	
*/

void quit(){
	printf("\nCTRL-C....killing childs..\n");
	kill(pid1,SIGINT);
	if(pid2!=0)kill(pid2,SIGINT);
	//exit(1);
	//return;
}


void quit_child(){
	TRIPP_DEBUG(DEBUG,"Child exiting...");
	exit(1);
}



void usage(char *appname){
	printf("TRIPP ver: %s\n  usage:\n\t%s -o <interface> [-f <rulesfile>] [-S] [-g <gateway>]\n\n",VERSION,appname);
	exit(0);
}


//conta i bit a uno in countnm
int countbits(u_int32_t countnm){
  int nmbits;
  
  	for (nmbits = 0; countnm; countnm >>= 1) nmbits += countnm & 1;
  	
  	return nmbits;
}




int main(int argc, char **argv){

	char 	*inintf=NULL,
			*outintf=NULL;
			
	int 	mode=-1,
			singproc=0,
			c;
	
	//parameter for first process
	struct thparam par1;
	//parameter for second process
	struct thparam par2;
	
	//tunnel for first proc
	struct tunnel tun1;
	//tun for second proc
	struct tunnel tun2;
	
	//rulesfile name
    char *filename=NULL;
    
	//gateway ip
    char *gwip=NULL;
    int err=0;
    
    //netaddr in str (ex: 10.0.0.0/24)
    char mynet_str[18+1];
     
    extern char *optarg;
     
     
     
	if(argc==1)usage(argv[0]);
	
     //process commandline options and parameters
     while ((c = getopt(argc, argv, ":So:f:g:")) != -1) {
       switch(c) {
         case 'S':
         	singproc=1;     
            break;
         case 'g':
            gwip = optarg;  
            break;
         case 'f':
            filename = optarg;  
            break;
         case 'i':
         	inintf = optarg;
			break;
		 case 'o':
			outintf = optarg;
			break;
         case ':':
            printf("error: -%c: argument is mandatory \n", optopt);
            usage(argv[0]);
            break;
         case '?':
            printf("error: unknown option %c\n", optopt);
            usage(argv[0]);
            break;
       }
     }
	
	//if(mode==LIBNET_RAW4)inintf=outintf;
	mode=LIBNET_LINK;
	
	//if(inintf==NULL)
	inintf=DEFAULT_INIF;
	
	if(mode==-1 || outintf==NULL)usage(argv[0]);
	
	
	//set interfaces default values and adjust mtus
	init_intfs(inintf,outintf,&tun1);
 	tun1.mode=mode;
	par1.tun=&tun1;
	par1.singproc=singproc;
	
	if(gwip!=NULL){
		par1.gw=ip_to_int(gwip,&err);
		if(err)INIT_ERROR("parsing error:%s",gwip);
	}
	else par1.gw=0;
	
	
	if(filename!=NULL)
		strcpy(par1.frules,filename);
	else
		strcpy(par1.frules,"");
		
		
		
	init_intfs(outintf,outintf,&tun2);
	strcpy(par2.frules,"");
	tun2.mode=LIBNET_RAW4;
	par2.tun=&tun2;
	
	
	//catch CTRL-C and kill childs
	signal(SIGINT,quit);
	
	sprintf(mynet_str,"%s/%u",
  					int_to_ip(tun1.outif.ipaddr & tun1.outif.netmask),//network address
  					countbits(tun1.outif.netmask)//netmask bits
  					); 
	
	
	//add routing
  	route(mynet_str,tun1.outif.name,tun1.inif.name);
  	if( gwip != NULL)	
  		route_defgw(gwip,"127.0.0.1",tun1.outif.name,tun1.inif.name);
  	TRIPP_DEBUG(DEBUG,"routing table modifyed");
	
	if(!singproc){
	  //add firewall rule
  	  frw_deny(mynet_str,tun2.outif.name);
	  TRIPP_DEBUG(DEBUG,"firewall rule added");
	}
	
	
	//struct par is used for future thread implementation
	if(!(pid1=fork())){
		tripp_process(&par1);
	
	}

	if(!singproc){	
		sleep(1);
		if(!(pid2=fork())){
			tripp_process(&par2);
		}
		
		//wait for processes termination
		waitpid(pid2,NULL,0);
		frw_delete(mynet_str,tun2.outif.name);
		TRIPP_DEBUG(DEBUG,"firewall rule deleted");
		kill(pid2,SIGINT);
		kill(pid1,SIGINT);
	}
		
	waitpid(pid1,NULL,0);
	if(restoremtu(&(tun1.inif)))TRIPP_DEBUG(DEBUG,"%s MTU restored ",tun1.inif.name);
	if(restoremtu(&(tun1.outif)))TRIPP_DEBUG(DEBUG,"%s MTU restored ",tun1.outif.name);
	
	//delete routing
  	route(mynet_str,tun1.inif.name,tun1.outif.name);
  	if( gwip != NULL)	
  		route_defgw("127.0.0.1",gwip,tun1.inif.name,tun1.outif.name);
	TRIPP_DEBUG(DEBUG,"routing table restored");
	

	rmsocket();
	printf("Exiting..\n");
	exit(1);
	
}



int rprint(struct tqhead_f* tqp){
	struct filter *fp;
	//struct tqhead_f *tqp;

	//tqp = &filtertq;	
	
	for (fp = TAILQ_FIRST(tqp); fp != NULL; fp = TAILQ_NEXT(fp, next))
		printf("%d\n",fp->id);
		
	return 0;
	
}



void* tripp_process(void *vpar){
	//ioctl parameter
	int n=1;
	
	//pcap handle
	pcap_t *pcap_hnd;
	//pcap errbuff
 	char p_errbuf[PCAP_ERRBUF_SIZE];
 	
 	//pcapp filter
 	struct bpf_program filter;
	bpf_u_int32 mask;
	bpf_u_int32 net;
	//buffer for filterstring
 	u_char filterstring[500];
 	
 	//libnet handle
	libnet_t *libnet_hnd;
	//libnet errbuff
	char ln_errbuf[LIBNET_ERRBUF_SIZE];

	//struct for reading l2addr with libnet
	struct libnet_ether_addr *hwaddr;
	
	//parameter for pcap callbak
	struct pcap_parameter pcp;
	
	//tunnel
	struct tunnel *tun;
	
	//IPC socket
	int sok;
	
	//casted parameter for this function
	struct thparam* par=(struct thparam*)vpar;
	
	//rules list head and its pointer
	struct tqhead_f filtertq;
	struct tqhead_f *filtertqp=&filtertq;
	

	//tcp channels list head and its pointer
	struct tqhead_c tcpchantq;
	struct tqhead_c *tcpchantqp=&tcpchantq;
	
	//packets list head and its pointer
	struct tqhead_p pkttq;
	struct tqhead_p *pkttqp=&pkttq;
	
	
	tun=par->tun;
	
	//initialize and load rules table
	TAILQ_INIT(&filtertq);
	TAILQ_INIT(&tcpchantq);
	TAILQ_INIT(&pkttq);
	
	if(strcmp(par->frules,""))	
 		load_rules_table(par->frules,&filtertq);
	
 	
    //catch ctrl-c
	signal(SIGINT,quit_child);

    //init libnet
	libnet_hnd = libnet_init(tun->mode,tun->outif.name,ln_errbuf);
	if(libnet_hnd==NULL)
		INIT_ERROR("libnet init failed: %s\n",ln_errbuf);
	

  sprintf(filterstring,"udp or tcp and net %s/%u",
  					int_to_ip(tun->outif.ipaddr & tun->outif.netmask),//network address
  					countbits(tun->outif.netmask)//netmask bits
  					); 
 
  

  if(tun->mode == LIBNET_LINK){
    //set outif layer2 address
  	tun->outif.l2addr_size=6;
  	hwaddr=libnet_get_hwaddr(libnet_hnd);
  	memcpy(&(tun->outif).l2addr,&hwaddr->ether_addr_octet,6); 	
 	//init arp table
 	arpinit();
 	//load server-side IPC socket
 	if(!par->singproc)sok=ipc_server();
 	sprintf(filterstring,"%s and dst host not %s",filterstring,int_to_ip(tun->outif.ipaddr));
  }else{
  	//load client-side IPC socket
  	sok=ipc_connect();
  	//modify filterstring 
  	sprintf(filterstring,"%s and dst host %s",filterstring,int_to_ip(tun->outif.ipaddr));
  }
  
  //pcap_lookupnet(tun->inif.name, &net, &mask, p_errbuf);
  pcap_hnd=pcap_open_live(tun->inif.name, DEFAULT_SNAPLEN, 1, 0, p_errbuf);
  if(pcap_hnd==NULL)
	INIT_ERROR("pcap error:%s\n",p_errbuf);
	
  
  #ifdef BSD
   ioctl(pcap_fileno(pcap_hnd), BIOCIMMEDIATE, &n);
  #endif
 //compile pcap filter
 //if((pcap_compile(pcap_hnd, &filter, filterstring, 0,mask))==-1) 
 if((pcap_compile(pcap_hnd, &filter, filterstring, 0,tun->inif.netmask))==-1)
  	INIT_ERROR("%s\n","pcap filter error\n");
 //set the filter 	
 pcap_setfilter(pcap_hnd, &filter);
 
 
 // print inital output
 //printf("sniffing \"%s\" on %s \n",filterstring,tun->inif.name);
 if(tun->mode == LIBNET_LINK)
 	printf("rewriter for outgoing traffic started\n");
 else
 	printf("rewriter for incoming traffic started\n");
 
 //pcap parameter
 pcp.libnet_h=libnet_hnd;
 pcp.tun=par->tun;
 pcp.ipc_sok=sok;
 pcp.tqp=&filtertqp;
 pcp.tqc=&tcpchantqp;
 pcp.tqpk=&pkttqp;
 pcp.gw=par->gw;
 //start sniffing
 pcap_loop(pcap_hnd, 0, pcap_callback, (u_char*)&pcp);

//never reached..
 pcap_close(pcap_hnd);
  
 //return 1;
  
}

//frees pakcet list cell
void free_pklcell(struct pktlist* pl){
    free(pl->p);
    free(pl);
 }
 
 
void pcap_callback(u_char *param, const struct pcap_pkthdr *header, const u_char *pkt_data){
	//casted parameter for this function
	struct pcap_parameter *pcp=(struct pcap_parameter *) param;

	//new ip packet
	struct ip_packet *newpkt;
	
	//pinter to read packes list
	struct pktlist *pkt;
	
	//clone{n} counter
	int a;
	
	
	if(pcp->tun->mode==LIBNET_RAW4)
		ipc_getrule_ack(pcp->ipc_sok,*pcp->tqp,*pcp->tqc);
	 
	newpkt=(struct ip_packet*)malloc(sizeof(struct ip_packet));
	
	if( (load_ip_packet(pkt_data,newpkt,header->caplen,pcp->tun->inif.type,0)) == 0 ){ 
    	free(newpkt);
		return;
	}
	
	newpkt->fragsize=pcp->tun->mtu;

 	process_packet(newpkt,pcp->ipc_sok,*pcp->tqp,*pcp->tqc);
 	
 	/*
 	if(newpkt->fragsize!=pcp->tun->mtu){
 		printf("buffer overflow %d\n",newpkt->fragsize);
 		quit();
 	}
 	*/
 	
 	fragment_packet(*pcp->tqpk,newpkt,newpkt->fragsize,pcp->tun->inif.type);
 	
 	for (pkt = TAILQ_FIRST(*pcp->tqpk); pkt != NULL; pkt = TAILQ_NEXT(pkt, next)){
 	
		if(pkt->p->counter <= 0) goto close;
		//printf("pkt\n");
	
		if(!build_ip(pkt->p,pcp->libnet_h,pcp->tun->mtu)){
			RT_ERROR("UNEXPECTED ERROR 2\n"); 
			goto close;
		}
	
		if(pcp->tun->mode == LIBNET_LINK)
			if(!build_layer2(pkt->p,pcp->libnet_h,pcp->tun,pcp->gw)) goto close;
		
		for(a=0; a < pkt->p->counter; a++){
			if(libnet_write(pcp->libnet_h)==-1)
				RT_ERROR("Write error: %s\n", libnet_geterror(pcp->libnet_h));
		 }	
		
	close:
		libnet_clear_packet(pcp->libnet_h);
		free_pklcell(pkt);
		TAILQ_REMOVE(*pcp->tqpk,pkt,next);		
		

    }
    
}
















