/* 	
	tripp.h 	part of TRIPP Traffic Rewriter for IP (v4) Packets
	by poplix@papuasia.org 2005

*/

#include "config.h"

/*set it to 1 to display debug messages otherwise set it to 0*/
//#define DEBUG 1

/*set it to 1 to display warnings otherwise set it to 0*/
#define WARNINGS 1


#define _BSD_SOURCE 1

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/wait.h>
//param.h defines BSD in bsd systems 
#include <sys/param.h>
#include <libnet.h>

//#include <netinet/if_ether.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#ifdef BSD
 #include <net/bpf.h>
#endif

#include <pcap.h>

//#include <netdb.h> //!!!!!!
#include <net/if.h>
//#include <netinet/in.h>//|||||||||||||||
#include <netinet/in_systm.h>
#include <netinet/tcp.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>//||||!!!!!

#include <limits.h>
#include <errno.h>

//#include <pthread.h>

#include "queue.h"

#define BALIGN 0
#define MAXMTU 1500 
#define MAXL2HDRSIZE 14
#define MAXBUFFSIZE MAXMTU + MAXL2HDRSIZE + BALIGN
#define MAX_L2ADDR_SIZE 6
#define DEFAULT_SNAPLEN MAXBUFFSIZE



//nota.. il fatto che IP sia compreso tra 1 e 5 TCP tra 6 e 12 ecc.. influisce su filter_getl4()
//ip_TOS e IP_TTL son gia definiti da libnet.h ....
#define IPTOS	1
#define IP_ID	2
#define IPTTL	3
#define IP_SRC	4
#define IP_DST	5

#define TCP_SPORT	6
#define TCP_DPORT	7
#define TCP_SEQ		8
#define TCP_ACK		9
#define TCP_FLAGS	10
#define TCP_WIN		11
#define TCP_URP		12

#define UDP_SPORT		13
#define UDP_DPORT		14

#define FORCE_LAYER4 15

#define TCP_OPTIONS	16

#define IP_FLAGS_MF 17
#define IP_FLAGS_DF 18
#define IP_FLAGS_RF 19


#define PAYLOAD 30
#define PAYLOADSIZE 31
#define PAYLOAD_REWRITE 32

#define ACT_SET		1
#define ACT_CLONE	2
#define ACT_DROP	3
#define ACT_DEELAY	4
#define ACT_PREWRITE	5
#define ACT_FRAGMENT	6

#define COND_MATCH 1

#define OPERATOR_EXACT 1
#define OPERATOR_CONTAINS 2
#define OPERATOR_DECREASE 3
#define OPERATOR_INCREASE 4
#define OPERATOR_LESSTHEN 5
#define OPERATOR_GREATERTHEN 6


#define TCPOPT_EOOL 	0
#define TCPOPT_NOOP		1
#define TCPOPT_MSS		2
#define TCPOPT_WSS		3
#define TCPOPT_SACKP	4
#define TCPOPT_TS		8


struct tcpopt{
	char ord[10];
	u_char *eool;
	u_char *noop;
	u_char *mss;
	int 	mss_size;
	u_char *wss;
	int 	wss_size;
	u_char *ts;
	int 	ts_size;
};


TAILQ_HEAD(tqhead_f, filter);
TAILQ_HEAD(tqhead_c, tcpchan);
TAILQ_HEAD(tqhead_p, pktlist);

struct tcpchan{
	u_int32_t saddr;
	u_int32_t daddr;
	u_short sport;
	u_short dport;
	u_int ruleid;
	
	TAILQ_ENTRY(tcpchan) next;

};


struct pcap_parameter{
	struct tunnel* tun;
	int ipc_sok;
	struct tqhead_f **tqp;
	struct tqhead_c **tqc;
	struct tqhead_p **tqpk;
	
	u_int32_t gw;
	libnet_t *libnet_h;
};



struct intf{
	char name[10];
	u_int mtu;
	u_int oldmtu;
	u_int type;
	
	u_int32_t ipaddr;
	u_int32_t netmask;
	
	u_char  l2addr[MAX_L2ADDR_SIZE];
	u_int	l2addr_size;
};




struct tunnel{
	struct intf inif;
	struct intf outif;
	int mode;
	u_int mtu;
};


struct actions{
	u_char opr;
	
	u_char rtype;
	u_char rplc[1024*3];
	u_int  rplc_size;
	
	//valori usati solo con payload rewrite
	u_char matchpayload[1024*3];
	u_int  replacetimes;
	u_int  matchpayload_size;
	
	TAILQ_ENTRY(actions) next;
};


struct matches{
	u_char opr;
	u_int  opr_not;
	
	u_char type;
	u_char value[1024*3];///ACHTUNG!!!!
	u_int  value_size;
	
	TAILQ_ENTRY(matches) next;
};


struct filter{
	u_int id;
	u_int layer4;
	//actcmd==ACT_CLONE,ACT_DEELAY
	u_char actcmd;
	//usato da clone{n} e fragment{n}
	u_int intval;
	
	//actcmd==ACT_SET
	TAILQ_HEAD(tqhead_m, matches) matchestq;
	TAILQ_HEAD(tqhead_a, actions) actionstq;

	TAILQ_ENTRY(filter) next;
	
};


//from net/ethernet.h
struct  ether_header {
        u_char  ether_dhost[ETHER_ADDR_LEN];
        u_char  ether_shost[ETHER_ADDR_LEN];
        u_short ether_type;
};





//ho scelto un buffer statico per evitare di usare realloc() per pochi bytes
//questa decisione puo essere rivista seriamente
struct ip_packet{ 
	u_int  headers_size;
	struct ip *ip;
	u_char *ip_options;
	u_int  ip_options_size;
	
	u_char *layer4;
	u_char *layer4_options;
	u_int  layer4_options_size;
	
	u_char *payload; 
	u_int  payload_size;
	
	u_char raw[(MAXBUFFSIZE)*10]; //** !!! #@  
	
	// counter=0 => drop
	int counter;
	int deelay;
	//nota che nei ipfrags a fragesize non viene assegnato alcun val => contine ????
	int fragsize;
};


typedef struct null_packet{
	u_int nullframe;
	struct ip_packet *p;
}null_packet;


typedef struct eth_packet{
	struct ether_header *ethernet;
	struct ip_packet *p;
}eth_packet;


struct arpentry{
	u_int32_t ip;
	char mac[6];
	TAILQ_ENTRY(arpentry) next;
};


struct pktlist{
	struct ip_packet* p;
	
	TAILQ_ENTRY(pktlist) next;

};



void pcap_callback(u_char *param, const struct pcap_pkthdr *header, const u_char *pkt_data);

int build_tcp(struct ip_packet*, libnet_t*);

u_int setmtu(struct intf* ,u_int);
int getifinfo(char *,struct intf *);
u_int restoremtu(struct intf* ifi);

char *int_to_ip(u_int32_t ip);
char* arp_resolve(u_int32_t ipaddr,libnet_t* l,struct tunnel *tun);

char *arpget(u_int32_t ip);
int arpadd(u_int32_t ip, char* mac);
void arpinit();
int arpprint();

//int filter_action(u_char pk_type,u_char *value,struct ip_packet*);

struct filter *load_filter(char **,char **,u_char);
int ruleadd(char **,struct tqhead_f *);

int load_rules_table(char *filename,struct tqhead_f *);

int memreplace(char *str, u_int str_size, char* rplc, u_int rplc_size, char *buff, u_int buff_size,int limit);


struct tcpchan *gettcpchan(u_int32_t saddr,u_int32_t daddr,u_short sport,u_short dport,struct tqhead_c *tqc);
struct tcpchan *newtcpchan(u_int32_t saddr,u_int32_t daddr,u_short sport,u_short dport,u_int ruleid);
struct filter *getrule(u_int index,struct tqhead_f* tqp);


int frw_deny(char *ip,char *intf);
int frw_delete(char *ip,char *intf);
 
 
u_int ip_to_int(u_char *,u_int *);
char *int_to_ip(u_int32_t );


void print_error(int errcode,char *fn,char *message, ...);
void print_debug(char *fn,char *message, ...);
void print_warning(char *fn,char *message, ...);

int route(char *net,char *oldif,char *newif);
int route_defgw(char *oldgw,char *newgw,char *oldif,char *newif);

u_char filter_getl4(u_char ft);//!!! no char


int rmsocket();

int fragment_packet(struct tqhead_p *pkl_h,struct ip_packet* pkt,u_int mtu,int linktype);


